"""Wrapper for the coordinate descent ("QuickSearch")"""
import numpy as np

from pyhealth.calib.utils import one_hot_np

_CYTHON_ENABLED = False
try:
    import pyximport
    pyximport.install()
    from . import quicksearch_cython as cdc
    _CYTHON_ENABLED = True
except:
    print("This is a warning of potentially slow compute. You could uncomment this line and use the Python implementation instead of Cython.")

__all__ = ['loss_overall', 'loss_classspecific',
           'coord_desc_overall', 'coord_desc_classspecific']

def _thresholding_py(ts, output):
    pred = np.asarray(output > ts, dtype=np.int32)
    return pred

def __loss_overall_helper(total_err, total_sure, alpha, N, la, lc, lcs):
    ambiguity_loss = (1. - total_sure / float(N)) ** 2
    risk = total_err / float(max(total_sure, 1))
    tempf = risk - alpha
    coverage_loss = np.power(max(tempf, 0), 2)
    coverage_loss_sim = np.power(tempf, 2)
    loss = ambiguity_loss * la + coverage_loss * lc + coverage_loss_sim * lcs
    return loss

def __loss_class_specific_complete_helper(err, sure, rs, weights, total_sure, N, la, lc, lcs):
    a_loss = 1- total_sure / N
    if sure.min() == 0: return np.inf
    tempf = err / sure
    if rs is not None:
        tempf -= rs
    if weights is None:
        c_loss = np.power(tempf.clip(0, 1), 2).sum()
        cs_loss = np.power(tempf, 2).sum()
    else:
        c_loss = np.dot(np.power(tempf.clip(0, 1), 2), weights)
        cs_loss = np.dot(np.power(tempf, 2), weights)
    return a_loss * la + c_loss * lc + cs_loss * lcs

def __update_cnts(pred, y, err, sure, inc=1):
    if pred != y:
        err[y] += inc
    sure[y] += inc
    return inc

def loss_overall_py(preds, labels, max_classes, risk, *, lk=1e4, fill_max=False):
    cnt = preds.sum(1)
    cnt1 = np.expand_dims(cnt == 1, 1)
    total_sure = sum(cnt1.squeeze(1))
    risk_indicator = labels * (1 - preds)
    total_err = (risk_indicator * cnt1).sum()
    if fill_max:
        cnt0 = cnt == 0
        total_sure += sum(cnt0)
        new_risk = max_classes != np.argmax(labels, 1)
        total_err += sum(cnt0 & new_risk)
    return __loss_overall_helper(total_err, total_sure, risk, len(labels), la=1, lc=lk, lcs=0)

def loss_class_specific_py(preds, labels, max_classes, risk, *, class_weights=False, lk=1e4, fill_max=False):
    N,K = preds.shape
    cnt = preds.sum(1)
    sure_msk = cnt == 1
    total_sure = np.sum(sure_msk)
    correct_msk = np.sum(preds * labels, 1) == 1
    sure = np.sum(labels[sure_msk], 0)
    err = np.sum(labels[sure_msk & ~correct_msk], 0)
    if fill_max:
        sure_msk = cnt == 0
        total_sure += np.sum(sure_msk)
        correct_msk = max_classes == np.argmax(labels, 1)
        sure += np.sum(labels[sure_msk], 0)
        err += np.sum(labels[sure_msk & ~correct_msk], 0)
    if isinstance(class_weights, bool):
        if class_weights:
            class_weights = np.asarray(
                np.unique(labels, return_counts=True)[1], dtype=np.float64
                ) * K / float(N)
        else:
            class_weights = None
    elif class_weights is not None:
        class_weights = np.asarray(class_weights)
    return __loss_class_specific_complete_helper(
        err, sure, risk, class_weights, total_sure, N, la=1, lc=lk, lcs=0)


def search_full_class_specific_py(mo, rnkscores_or_maxclasses, scores_idx, labels, alphas, class_weights, ps, d, *,
                                  lk=1e4, fill_max=False):
    # d is dimesion to search/descend
    N,K = mo.shape
    ps = ps.copy()
    if len(rnkscores_or_maxclasses.shape) == 2:
        ts = [(rnkscores_or_maxclasses[ps[ki], ki]) for ki in range(K)]
        max_classes = np.argmax(mo, 1)
    else:
        ts = [ps[ki] for ki in range(K)]
        max_classes = rnkscores_or_maxclasses

    preds = _thresholding_py(ts, mo)
    preds[:, d] = 1
    cnt = preds.sum(1)
    labels_onehot = one_hot_np(labels, K)
    cnt1_msk = cnt == 1
    correct_msk = cnt1_msk & (np.sum(preds * labels_onehot, 1) == 1)
    total_sure = np.sum(cnt1_msk)
    sure = np.sum(labels_onehot[cnt1_msk], 0)
    err = np.sum(labels_onehot[cnt1_msk & ~correct_msk], 0)
    if fill_max:
        cnt0_msk = cnt == 0
        cnt0_correct_msk = max_classes == labels
        total_sure += np.sum(cnt0_msk)
        sure += np.sum(labels_onehot[cnt0_msk], 0)
        err += np.sum(labels_onehot[cnt0_msk & ~cnt0_correct_msk], 0)
    _preds = np.ones(N, dtype=int) * K
    _preds[cnt1_msk] = np.argmax(preds[cnt1_msk], 1)
    cnt2_msk = np.asarray((cnt == 2) & (preds[:, d]), dtype=bool) #second mask is trivial
    preds[:, d] = 0 #this order cannot chance
    _preds[cnt2_msk] = np.argmax(preds[cnt2_msk], 1)

    best_i, best_loss = -1, np.inf
    for ijjj in range(N-1):
        i = scores_idx[ijjj, d]
        yi = labels[i]
        tint = _preds[i]
        if cnt[i] == 1 or cnt[i] == 2:
            total_sure += __update_cnts(tint, yi, err, sure, 1 if cnt[i] == 2 else -1)
            if fill_max and cnt[i] == 1:
                total_sure += __update_cnts(max_classes[i], yi, err, sure, 1)
        cnt[i] -= 1
        curr_loss = __loss_class_specific_complete_helper(
            err, sure, alphas, class_weights, total_sure, N, la=1, lc=lk, lcs=0)
        if curr_loss < best_loss:
            best_i, best_loss = ijjj, curr_loss
    return best_i, best_loss


def search_full_overall_py(mo, rnkscores_or_maxclasses, scores_idx, labels, alpha, ps, d, *,
                           lk=1e4, fill_max=False):
    N,K = mo.shape
    ps = ps.copy()
    if len(rnkscores_or_maxclasses.shape) == 2:
        ts = [(rnkscores_or_maxclasses[ps[ki], ki]) for ki in range(K)]
        max_classes = np.argmax(mo, 1)
    else:
        ts = [ps[ki] for ki in range(K)]
        max_classes = rnkscores_or_maxclasses
    preds = _thresholding_py(ts, mo)
    preds[:, d] = 1
    cnt = preds.sum(1)

    #initialize mems for ijjj=0
    total_err, total_sure = 0, 0
    for ii in range(N):
        yii = labels[ii]
        if cnt[ii] == 1:
            total_sure += 1
            total_err += 1 - preds[ii, yii]
        if cnt[ii] == 0 and fill_max:
            total_sure += 1
            if max_classes[ii] != yii:
                total_err += 1
    best_i, best_loss = -1, np.inf

    for ijjj in range(N-1):
        i = scores_idx[ijjj, d]
        yi = labels[i]
        #preds[i, d] will change from 1 to 0
        if cnt[i] == 2:#unsure -> sure
            if d == yi:#we miss this item in this case, so error increases by 1
                total_err += 1
            elif mo[i, yi] <= ts[yi]: #we miss the item
                total_err += 1
            total_sure += 1
        elif cnt[i] == 1: #Sure -> unsure. Also, this cnt has to be class d
            if d != yi:
                total_err -= 1
            total_sure -= 1
            if fill_max:
                total_sure += 1
                if max_classes[i] != yi:
                    total_err += 1
        else: #Do nothing as we have an unsure -> unsure case
            pass
        cnt[i] -= 1

        curr_loss = __loss_overall_helper(total_err, total_sure, alpha, N, la=1, lc=lk, lcs=0)

        if curr_loss < best_loss:
            best_i, best_loss = ijjj, curr_loss
    return best_i, best_loss

def coord_desc_classspecific_py(mo, rnkscores_or_maxclasses, scores_idx, labels, ps, alphas, *,
                                          class_weights=False, lk=1e4, fill_max=False):
    N, K = mo.shape
    best_loss, ps = np.inf, ps.copy()

    keep_going = True
    if isinstance(class_weights, bool):
        if class_weights:
            weights = np.sum(one_hot_np(labels, K), 0) / float(N) * K
        else:
            weights = None
    else:
        weights = class_weights

    while keep_going:
        curr_ps = np.zeros(K)
        best_ki, curr_loss = None, best_loss
        for ki in range(K):
            curr_ps[ki], temp_loss = search_full_class_specific_py(
                mo, rnkscores_or_maxclasses, scores_idx, labels, alphas, weights, ps, ki,
                lk=lk, fill_max=fill_max)
            if temp_loss < curr_loss:
                best_ki, curr_loss = ki, temp_loss
        if curr_loss < best_loss:
            ps[best_ki] = curr_ps[best_ki]
            best_loss = curr_loss
        else:
            keep_going = False
    return best_loss, ps, None


def coord_desc_overall_py(mo, rnkscores_or_maxclasses, scores_idx, labels, ps, alpha, *,
                                   lk=1e4, fill_max=False):
    N, K = mo.shape
    best_loss, ps = np.inf, ps.copy()

    keep_going = True

    while keep_going:
        curr_ps = np.zeros(K)
        best_ki, curr_loss = None, best_loss
        for ki in range(K):
            curr_ps[ki], temp_loss = search_full_overall_py(
                mo, rnkscores_or_maxclasses, scores_idx, labels, alpha, ps, ki,
                lk=lk, fill_max=fill_max)
            if temp_loss < curr_loss:
                best_ki, curr_loss = ki, temp_loss
        if curr_loss < best_loss:
            ps[best_ki] = curr_ps[best_ki]
            best_loss = curr_loss
        else:
            keep_going = False
    return best_loss, ps, None

def loss_overall(idx2rnk, rnk2idx, labels, max_classes, ps, r, *,
                 lk=1e4, fill_max=False):
    if not _CYTHON_ENABLED:
        preds = np.asarray(idx2rnk > ps, np.int32)
        return loss_overall_py(
            preds, one_hot_np(labels, idx2rnk.shape[1]), max_classes, r,
            lk=lk, fill_max=fill_max)
    idx2rnk = np.asarray(idx2rnk, np.int32)
    rnk2idx = np.asarray(rnk2idx, np.int32)
    labels = np.asarray(labels, np.int32)
    max_classes = np.asarray(max_classes, np.int32)
    ps = np.asarray(ps, np.int32)
    return cdc.loss_overall_c(
        idx2rnk, rnk2idx, labels, max_classes, ps, r,
        la=1, lc=lk, lcs=0, fill_max=fill_max)

def loss_classspecific(idx2rnk, rnk2idx, labels, max_classes, ps, rks, *,
                        class_weights=None, lk=1e4, fill_max=False):
    if not _CYTHON_ENABLED:
        preds = np.asarray(idx2rnk > ps, np.int32)
        return loss_class_specific_py(
            preds, one_hot_np(labels, idx2rnk.shape[1]), max_classes, rks,
            class_weights=class_weights, lk=lk, fill_max=fill_max)
    idx2rnk = np.asarray(idx2rnk, np.int32)
    rnk2idx = np.asarray(rnk2idx, np.int32)
    labels = np.asarray(labels, np.int32)
    max_classes = np.asarray(max_classes, np.int32)
    ps = np.asarray(ps, np.int32)
    if class_weights is not None:
        class_weights = np.asarray(class_weights, np.float64)
    if rks is not None:
        rks = np.asarray(rks, np.float64)
    return cdc.loss_class_specific_c(
        idx2rnk, rnk2idx, labels, max_classes, ps, rks,
        class_weights, la=1, lc=lk, lcs=0, fill_max=fill_max)

def coord_desc_overall(idx2rnk, rnk2idx, labels, max_classes, init_ps, r, *,
                               max_step_size=None, lk=1e4, fill_max=False):
    if not _CYTHON_ENABLED:
        assert max_step_size is None
        return coord_desc_overall_py(
            idx2rnk, max_classes, rnk2idx, labels, init_ps, r,
            lk=lk, fill_max=fill_max)
    idx2rnk = np.asarray(idx2rnk, np.int32)
    rnk2idx = np.asarray(rnk2idx, np.int32)
    labels = np.asarray(labels, np.int32)
    max_classes = np.asarray(max_classes, np.int32)
    init_ps = np.asarray(init_ps, np.int32)
    return cdc.coord_desc_overall_c(
        idx2rnk, rnk2idx, labels, max_classes, init_ps, r,
        max_step_size, la=1, lc=lk, lcs=0, fill_max=fill_max)


def coord_desc_classspecific(idx2rnk, rnk2idx, labels, max_classes, init_ps, rks, *,
                                      class_weights=None,
                                      max_step_size=None, lk=1e4, fill_max=False):
    if not _CYTHON_ENABLED:
        assert max_step_size is None
        return coord_desc_classspecific_py(
            idx2rnk, max_classes, rnk2idx, labels, init_ps, rks,
            class_weights=class_weights, lk=lk, fill_max=fill_max)
    idx2rnk = np.asarray(idx2rnk, np.int32)
    rnk2idx = np.asarray(rnk2idx, np.int32)
    labels = np.asarray(labels, np.int32)
    max_classes = np.asarray(max_classes, np.int32)
    init_ps = np.asarray(init_ps, np.int32)
    if class_weights is not None:
        class_weights = np.asarray(class_weights, np.float64)
    if rks is not None:
        rks = np.asarray(rks, np.float64)
    return cdc.coord_desc_classspecific_c(
        idx2rnk, rnk2idx, labels, max_classes, init_ps, rks,
        class_weights, max_step_size, la=1, lc=lk, lcs=0, fill_max=fill_max)



"""
Fast Online Value-Maximizing Prediction Sets with Conformal Cost Control (FavMac)

Implementation based on https://github.com/zlin7/FavMac

"""

from importlib import reload
import time
from typing import Dict, Union

import numpy as np
import pandas as pd
import torch

from pyhealth.calib.base_classes import SetPredictor
from pyhealth.calib.predictionset.favmac.core import FavMac_GreedyRatio
from pyhealth.calib.utils import prepare_numpy_dataset
from pyhealth.models import BaseModel

__all__ = ["FavMac"]

INTEGER_SAFE_DELTA = 0.1


class AdditiveSetFunction:
    def __init__(self, values: Union[float, np.ndarray, int], mode=None, name='unknown') -> None:
        self.name = name
        self.values = values
        assert mode is None or mode in {'util', 'cost', 'proxy'}
        self.mode = mode

    def is_additive(self):
        return True

    def __call__(self, S: np.ndarray, Y:np.ndarray=None, pred:np.ndarray=None, sample=100, target_cost=None) -> float:
        if self.mode == 'cost':
            assert pred is None
            return self.cost_call(S, Y)
        if self.mode == 'proxy':
            assert Y is None
            return self.proxy_call(S, pred, target_cost=target_cost)
        assert self.mode == 'util'
        return self.util_call(S, Y, pred, sample=sample)

    def naive_call(self, S: np.ndarray) -> float:
        return np.sum(S * self.values)

    def util_call(self, S: np.ndarray, Y:np.ndarray=None, pred:np.ndarray=None, sample=1000) -> float:
        assert Y is None or pred is None
        if pred is not None:
            return self.naive_call(S * pred) # This is because of additivity.
        if Y is not None: return self.naive_call(S * Y)
        return self.naive_call(S)

    def cost_call(self, S: np.ndarray, Y:np.ndarray) -> float:
        return self.naive_call(S * (1-Y))

    def proxy_call(self, S: np.ndarray, pred: np.ndarray, target_cost: float=None) -> float:
        return self.naive_call(S * (1-pred))

    def greedy_maximize(self, S: np.ndarray, pred: np.ndarray=None, d_proxy:np.ndarray=None, prev_util_and_proxy=None):
        # (prev_u, prev_p) = prev_util_and_proxy
        assert self.mode == 'util', "This is only used for util function"
        if (1-S).sum() == 0: return None
        d_util = self.values
        if pred is not None: d_util = d_util * pred

        objective = d_util / (1 if d_proxy is None else d_proxy.clip(1e-8))
        k = pd.Series((1-S) * objective).dropna().idxmax()
        return k, objective[k]

    def greedy_maximize_seq(self, pred: np.ndarray=None, d_proxy:np.ndarray=None):
        # if cost is also additive, then cost_proxy is fixed: weight * (1-p)
        assert self.mode == 'util', "This is only used for util function"
        d_util = self.values
        if pred is not None: d_util = d_util * pred

        objective = d_util / (1 if d_proxy is None else d_proxy.clip(1e-8))

        assert np.isnan(objective).sum() == 0
        ks = np.argsort(-objective)
        Ss = [np.zeros(len(objective), dtype=int)]
        for k in ks:
            Ss.append(Ss[-1].copy())
            Ss[-1][k] = 1
        return Ss, ks


class FavMac(SetPredictor):
    """Fast Online Value-Maximizing Prediction Sets with Conformal Cost Control (FavMac)

    This is a prediction-set constructor for multi-label classification problems.
    FavMac could control the cost/risk while realizing high value on the prediction set.

    Value and cost functions are functions in the form of :math:`V(S;Y)` or :math:`C(S;Y)`,
    with S being the prediction set and Y being the label.
    For example, a classical cost function would be "numebr of false positives".
    Denote the ``target_cost`` as
    :math:`c`,
    if ``delta=None``, FavMac controls the expected cost in the following sense:

    :math:`\\mathbb{E}[C(S_{N+1};Y_{N+1}] \\leq c`.

    Otherwise, FavMac controls the violation probability in the following sense:

    :math:`\\mathbb{P}\\{C(S_{N+1};Y_{N+1})>c\\}\\leq delta`.

    Right now, this FavMac implementation only supports additive value and cost functions (unlike the
    implementation associated with [1]).
    That is, the value function is specified by the weights ``value_weights`` and the cost function
    is specified by ``cost_weights``.
    With :math:`k` denoting classes, the cost function is then computed as

    :math:`C(S;Y,w) = \\sum_{k} (1-Y_k)S_k w_k`

    Similarly, the value function is computed as

    :math:`V(S;Y,w) = \\sum_{k} Y_k S_k w_k`.

    Papers:

        [1] Lin, Zhen, Shubhendu Trivedi, Cao Xiao, and Jimeng Sun.
        "Fast Online Value-Maximizing Prediction Sets with Conformal Cost Control (FavMac)."
        ICML 2023.

        [2] Fisch, Adam, Tal Schuster, Tommi Jaakkola, and Regina Barzilay.
        "Conformal prediction sets with limited false positives."
        ICML 2022.

    Args:
        model (BaseModel): A trained model.
        value_weights (Union[float, np.ndarray]):
            weights for the value function. See description above.
            Defaults to 1.
        cost_weights (Union[float, np.ndarray]):
            weights for the cost function. See description above.
            Defaults to 1.
        target_cost (float): Target cost.
            When cost_weights is set to 1, this is essentially the number of false positive.
            Defaults to 1.
        delta (float): Violation target (in violation control).
            Defaults to None (which means expectation control instead of violation control).

    Examples:
        >>> from pyhealth.calib.predictionset import FavMac
        >>> from pyhealth.datasets import (MIMIC3Dataset, get_dataloader,split_by_patient)
        >>> from pyhealth.models import Transformer
        >>> from pyhealth.tasks import drug_recommendation_mimic3_fn
        >>> from pyhealth.trainer import get_metrics_fn
        >>> base_dataset = MIMIC3Dataset(
        ...     root="/srv/scratch1/data/physionet.org/files/mimiciii/1.4",
        ...     tables=["DIAGNOSES_ICD", "PROCEDURES_ICD", "PRESCRIPTIONS"],
        ...     code_mapping={"NDC": ("ATC", {"target_kwargs": {"level": 3}})},
        ...     refresh_cache=False)
        >>> sample_dataset = base_dataset.set_task(drug_recommendation_mimic3_fn)
        >>> train_data, val_data, test_data = split_by_patient(sample_dataset, [0.6, 0.2, 0.2])
        >>> model = Transformer(dataset=sample_dataset, feature_keys=["conditions", "procedures"],
        ...             label_key="drugs", mode="multilabel")
        >>> # ... Train the model here ...
        >>> # Try to control false positive to <=3
        >>> cal_model = FavMac(model, target_cost=3, delta=None)
        >>> cal_model.calibrate(cal_dataset=val_data)
        >>> # Evaluate
        >>> from pyhealth.trainer import Trainer
        >>> test_dl = get_dataloader(test_data, batch_size=32, shuffle=False)
        >>> y_true_all, y_prob_all, _, extra_output = Trainer(model=cal_model).inference(
        ... test_dl, additional_outputs=["y_predset"])
        >>> print(get_metrics_fn(cal_model.mode)(
        ...     y_true_all, y_prob_all, metrics=['tp', 'fp'],
        ...     y_predset=extra_output["y_predset"])) # We get FP~=3
        {'tp': 0.5049893086243763, 'fp': 2.8442622950819674}
    """

    def __init__(
        self,
        model: BaseModel,
        value_weights: Union[float, np.ndarray] = 1.,
        cost_weights: Union[float, np.ndarray] = 1.,
        target_cost: float = 1., delta:float = None,
        debug=False,
        **kwargs,
    ) -> None:
        super().__init__(model, **kwargs)
        if model.mode != "multilabel":
            raise NotImplementedError()
        self.mode = self.model.mode  # multilabel
        for param in model.parameters():
            param.requires_grad = False
        self.model.eval()
        self.device = model.device
        self.debug = debug


        self._cost_weights = cost_weights
        self._value_weights = value_weights
        self.target_cost = target_cost
        self.delta = delta

    def calibrate(self, cal_dataset):
        """Calibrate the cost-control procedure.

        :param cal_dataset: Calibration set.
        :type cal_dataset: Subset
        """
        _cal_data = prepare_numpy_dataset(
            self.model, cal_dataset, ["logit", "y_true"], debug=self.debug
        )

        if isinstance(self._cost_weights, np.ndarray):
            C_max = self._cost_weights.sum()
        else:
            C_max = _cal_data["logit"].shape[1] * self._cost_weights
        self._favmac = FavMac_GreedyRatio(
            cost_fn=AdditiveSetFunction(self._cost_weights / C_max, mode='cost'),
            util_fn=AdditiveSetFunction(self._value_weights, mode='util'),
            proxy_fn=AdditiveSetFunction(self._cost_weights / C_max, mode='proxy'),
            target_cost=self.target_cost/C_max, delta=self.delta, C_max=1.,
        )
        self._favmac.init_calibrate(_cal_data["logit"], _cal_data["y_true"])

    def forward(self, **kwargs) -> Dict[str, torch.Tensor]:
        """Forward propagation (just like the original model).

        :return: A dictionary with all results from the base model, with the following updates:

                y_predset: a bool tensor representing the prediction for each class.
        :rtype: Dict[str, torch.Tensor]
        """
        ret = self.model(**kwargs)
        _logit = ret["logit"].cpu().numpy()
        y_predset = np.asarray([self._favmac(_)[0] for _ in _logit])

        ret["y_predset"] = torch.tensor(y_predset)
        return ret


if __name__ == "__main__":
    from pyhealth.calib.predictionset import FavMac
    from pyhealth.datasets import (MIMIC3Dataset, get_dataloader,
                                   split_by_patient)
    from pyhealth.models import Transformer
    from pyhealth.tasks import drug_recommendation_mimic3_fn
    from pyhealth.trainer import get_metrics_fn

    base_dataset = MIMIC3Dataset(
        root="/srv/scratch1/data/physionet.org/files/mimiciii/1.4",
        tables=["DIAGNOSES_ICD", "PROCEDURES_ICD", "PRESCRIPTIONS"],
        code_mapping={"NDC": ("ATC", {"target_kwargs": {"level": 3}})},
        refresh_cache=False,
    )
    sample_dataset = base_dataset.set_task(drug_recommendation_mimic3_fn)
    train_data, val_data, test_data = split_by_patient(sample_dataset, [0.6, 0.2, 0.2])
    model = Transformer(dataset=sample_dataset, feature_keys=["conditions", "procedures"],
                        label_key="drugs", mode="multilabel")
    # ... Train the model here ...
    # calibrate the prediction sets with FavMac
    cal_model = FavMac(model, cost_weights=1., target_cost=3, delta=None)
    cal_model.calibrate(cal_dataset=val_data)
    # Evaluate
    from pyhealth.trainer import Trainer

    test_dl = get_dataloader(test_data, batch_size=32, shuffle=False)
    y_true_all, y_prob_all, _, extra_output = Trainer(model=cal_model).inference(
        test_dl, additional_outputs=["y_predset"]
    )
    print(
        get_metrics_fn(cal_model.mode)(
            y_true_all, y_prob_all, metrics=['tp', 'fp'],
            y_predset=extra_output["y_predset"],
        )
    )



"""Model calibration methods"""
from pyhealth.calib.calibration.dircal import DirichletCalibration
from pyhealth.calib.calibration.hb import HistogramBinning
from pyhealth.calib.calibration.kcal import KCal
from pyhealth.calib.calibration.temperature_scale import TemperatureScaling

__all__ = ['DirichletCalibration', 'HistogramBinning', 'KCal', 'TemperatureScaling']



import numpy as np

"""
Notation:
    - Protected group: P
    - Unprotected group: U
"""

def disparate_impact(sensitive_attributes: np.ndarray, y_pred: np.ndarray, favorable_outcome: int = 1, allow_zero_division = False, epsilon: float = 1e-8) -> float:
    """
    Computes the disparate impact between the the protected and unprotected group.

    disparate_impact = P(y_pred = favorable_outcome | P) / P(y_pred = favorable_outcome | U)
    
    Args:
        sensitive_attributes: Sensitive attributes of shape (n_samples,) where 1 is the protected group and 0 is the unprotected group.
        y_pred: Predicted target values of shape (n_samples,).
        favorable_outcome: Label value which is considered favorable (i.e. "positive").
        allow_zero_division: If True, use epsilon instead of 0 in the denominator if the denominator is 0. Otherwise, raise a ValueError.
    
    Returns:
        The disparate impact between the protected and unprotected group.
    """
    
    p_fav_unpr = np.sum(y_pred[sensitive_attributes == 0] == favorable_outcome) / len(y_pred[sensitive_attributes == 0])
    p_fav_prot = np.sum(y_pred[sensitive_attributes == 1] == favorable_outcome) / len(y_pred[sensitive_attributes == 1])

    if p_fav_unpr == 0:
        if allow_zero_division:
            p_fav_unpr = epsilon
        else:
            raise ValueError("Unprotected group has no instances with a favorable outcome. Disparate impact is undefined.")

    disparate_impact_value = p_fav_prot / p_fav_unpr

    return disparate_impact_value

def statistical_parity_difference(sensitive_attributes: np.ndarray, y_pred: np.ndarray, favorable_outcome: int = 1) -> float:
    """
    Computes the statistical parity difference between the the protected and unprotected group.

    statistical_parity_difference = P(y_pred = favorable_outcome | P) - P(y_pred = favorable_outcome | U)
    Args:
        sensitive_attributes: Sensitive attributes of shape (n_samples,) where 1 is the protected group and 0 is the unprotected group.
        y_pred: Predicted target values of shape (n_samples,).
        favorable_outcome: Label value which is considered favorable (i.e. "positive").
    Returns:
        The statistical parity difference between the protected and unprotected group.
    """

    p_fav_unpr = np.sum(y_pred[sensitive_attributes == 0] == favorable_outcome) / len(y_pred[sensitive_attributes == 0])
    p_fav_prot = np.sum(y_pred[sensitive_attributes == 1] == favorable_outcome) / len(y_pred[sensitive_attributes == 1])
    
    statistical_parity_difference_value = p_fav_prot - p_fav_unpr

    return statistical_parity_difference_value



    


"""
Recommended:
 - torch: 1.9.1
 - torchvision: 0.10.0+cu102
 - torch_sparse: 0.6.12
"""

from typing import List, Dict

import math
import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision

import sys
sys.path.append('.')

from pyhealth.datasets.sample_dataset_v2 import SampleDataset
from pyhealth.models import BaseModel
from pyhealth.sampler import NeighborSampler

SUPPORTED_MODELS = [
    "resnet18",
    "resnet34",
    "resnet50",
    "resnet101",
    "resnet152",
    "densenet121",
    "densenet161",
    "densenet169",
    "densenet201",
    "vit_b_16",
    "vit_b_32",
    "vit_l_16",
    "vit_l_32",
    "vit_h_14",
    "swin_t",
    "swin_s",
    "swin_b",
]

SUPPORTED_MODELS_FINAL_LAYER = {}
for model in SUPPORTED_MODELS:
    if "resnet" in model:
        SUPPORTED_MODELS_FINAL_LAYER[model] = "fc"
    elif "densenet" in model:
        SUPPORTED_MODELS_FINAL_LAYER[model] = "classifier"
    elif "vit" in model:
        SUPPORTED_MODELS_FINAL_LAYER[model] = "heads.head"
    elif "swin" in model:
        SUPPORTED_MODELS_FINAL_LAYER[model] = "head"
    else:
        raise NotImplementedError


class GraphConvolution(nn.Module):
    """
    Simple GCN layer, similar to https://arxiv.org/abs/1609.02907
    """

    def __init__(self, in_features, out_features, bias=True, init='xavier'):
        super(GraphConvolution, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = nn.Parameter(torch.FloatTensor(in_features, out_features))
        if bias:
            self.bias = nn.Parameter(torch.FloatTensor(out_features))
        else:
            self.register_parameter('bias', None)
        if init == 'uniform':
            print("| Uniform Initialization")
            self.reset_parameters_uniform()
        elif init == 'xavier':
            print("| Xavier Initialization")
            self.reset_parameters_xavier()
        elif init == 'kaiming':
            print("| Kaiming Initialization")
            self.reset_parameters_kaiming()
        else:
            raise NotImplementedError

    def reset_parameters_uniform(self):
        stdv = 1. / math.sqrt(self.weight.size(1))
        self.weight.data.uniform_(-stdv, stdv)
        if self.bias is not None:
            self.bias.data.uniform_(-stdv, stdv)

    def reset_parameters_xavier(self):
        nn.init.xavier_normal_(self.weight.data, gain=0.02) # Implement Xavier Uniform
        if self.bias is not None:
            nn.init.constant_(self.bias.data, 0.0)

    def reset_parameters_kaiming(self):
        nn.init.kaiming_normal_(self.weight.data, a=0, mode='fan_in')
        if self.bias is not None:
            nn.init.constant_(self.bias.data, 0.0)

    def forward(self, input, adj):
        support = torch.mm(input, self.weight)
        # print("adj", adj.dtype, "support", support.dtype)
        output = torch.spmm(adj, support)
        if self.bias is not None:
            return output + self.bias
        else:
            return output

    def __repr__(self):
        return self.__class__.__name__ + ' (' \
               + str(self.in_features) + ' -> ' \
               + str(self.out_features) + ')'


class GCN(nn.Module):
    def __init__(self, nfeat, nhid, nclass, dropout, init):
        super(GCN, self).__init__()

        self.gc1 = GraphConvolution(nfeat, nhid, init=init)
        self.gc2 = GraphConvolution(nhid, nclass, init=init)
        self.dropout = dropout

    def bottleneck(self, path1, path2, path3, adj, in_x):
        return F.relu(path3(F.relu(path2(F.relu(path1(in_x, adj)), adj)), adj))

    def to_sparse_adj(self, adj, size):
        return torch.sparse_coo_tensor(adj.edge_index, torch.ones_like(adj.edge_index[0]), size = size, dtype = torch.float32)

    def forward(self, x, adjs):
        
        temp = self.to_sparse_adj(adjs[0], size = (adjs[0].size[0], adjs[0].size[0]))
        x = F.dropout(F.relu(self.gc1(x, temp)), self.dropout, training=self.training)
        temp = self.to_sparse_adj(adjs[1], size = (adjs[0].size[0], adjs[0].size[0]))
        x = self.gc2(x, temp)

        return F.log_softmax(x, dim=1)


class Graph_TorchvisionModel(BaseModel):
    """Models from PyTorch's torchvision package.

    This class is a wrapper for models from torchvision. It will automatically load
    the corresponding model and weights from torchvision. The final layer will be
    replaced with a linear layer with the correct output size.

    -----------------------------------ResNet------------------------------------------
    Paper: Kaiming He, Xiangyu Zhang, Shaoqing Ren, Jian Sun. Deep Residual Learning
    for Image Recognition. CVPR 2016.
    -----------------------------------DenseNet----------------------------------------
    Paper: Gao Huang, Zhuang Liu, Laurens van der Maaten, Kilian Q. Weinberger.
    Densely Connected Convolutional Networks. CVPR 2017.
    ----------------------------Vision Transformer (ViT)-------------------------------
    Paper: Alexey Dosovitskiy, Lucas Beyer, Alexander Kolesnikov, Dirk Weissenborn, 
    Xiaohua Zhai, Thomas Unterthiner, Mostafa Dehghani, Matthias Minderer,
    Georg Heigold, Sylvain Gelly, Jakob Uszkoreit, Neil Houlsby. An Image is Worth
    16x16 Words: Transformers for Image Recognition at Scale. ICLR 2021.
    ----------------------------Swin Transformer (and V2)------------------------------
    Paper: Ze Liu, Yutong Lin, Yue Cao, Han Hu, Yixuan Wei, Zheng Zhang, Stephen Lin, 
    Baining Guo. Swin Transformer: Hierarchical Vision Transformer Using Shifted
    Windows. ICCV 2021.

    Paper: Ze Liu, Han Hu, Yutong Lin, Zhuliang Yao, Zhenda Xie, Yixuan Wei, Jia Ning, 
    Yue Cao, Zheng Zhang, Li Dong, Furu Wei, Baining Guo. Swin Transformer V2: Scaling
    Up Capacity and Resolution. CVPR 2022.
    -----------------------------------------------------------------------------------
    ----------------------- Graph Convolutional Networks (GCN)-------------------------
    Paper: Thomas N. Kipf, Max Welling. 
    Semi-Supervised Classification with Graph Convolutional Networks. ICLR 2017.
    -----------------------------------------------------------------------------------

    Args:
        dataset: the dataset to train the model. It is used to query certain
            information such as the set of all tokens.
        feature_keys:  list of keys in samples to use as features, e.g., ["image"].
            Only one feature is supported.
        label_key: key in samples to use as label, e.g., "drugs".
        mode: one of "binary", "multiclass", or "multilabel".
        model_name: str, name of the model to use, e.g., "resnet18".
            See SUPPORTED_MODELS in the source code for the full list.
        model_config: dict, kwargs to pass to the model constructor,
            e.g., {"weights": "DEFAULT"}. See the torchvision documentation for the
            set of supported kwargs for each model.
    -----------------------------------------------------------------------------------
    """

    def __init__(
        self,
        dataset: SampleDataset,
        feature_keys: List[str],
        label_key: str,
        mode: str,
        model_name: str,
        model_config: dict,
        gnn_config: dict,
    ):
        super(Graph_TorchvisionModel, self).__init__(
            dataset=dataset,
            feature_keys=feature_keys,
            label_key=label_key,
            mode=mode,
        )

        self.model_name = model_name
        self.model_config = model_config
        self.gnn_config = gnn_config

        assert len(feature_keys) == 1, "Only one feature is supported!"
        assert model_name in SUPPORTED_MODELS_FINAL_LAYER.keys(), \
            f"PyHealth does not currently include {model_name} model!"

        # for torchvision 0.10.0
        self.model = torchvision.models.__dict__[model_name](**model_config)
        final_layer_name = SUPPORTED_MODELS_FINAL_LAYER[model_name]
        final_layer = self.model
        for name in final_layer_name.split("."):
            final_layer = getattr(final_layer, name)
        hidden_dim = final_layer.in_features
        
        # Graph Model Configs
        gnn_input_dim = gnn_config["input_dim"]
        gnn_hidden_dim = gnn_config["hidden_dim"]

        self.label_tokenizer = self.get_label_tokenizer()
        output_size = self.get_output_size(self.label_tokenizer)
        self.gnn = GCN(nfeat=gnn_input_dim, nhid=gnn_hidden_dim, nclass=output_size, dropout=0.5, init='uniform')
        
        setattr(self.model, final_layer_name.split(".")[0], nn.Linear(hidden_dim, gnn_input_dim))


    def build_graph(self, data, random = False) -> Dict[str, torch.Tensor]:
        """This module generate edge index of graph structure based on given data.
        Currently, we do not have multi-modal data, so this module randomly generate edge index"""
        
        if random:
            edge_index = torch.randint(len(data), size = (2, int(1.0 * len(data))))

        return {
            'edge_index': edge_index
        }


    def forward(self, **kwargs) -> Dict[str, torch.Tensor]:
        """Forward propagation."""
        # concat the info within one batch (batch, channel, length)
        x = kwargs["image"]
        x = torch.stack(x, dim=0).to(self.device)
        if x.shape[1] == 1:
            x = x.repeat((1, 3, 1, 1))
        img_embs = self.model(x)
        logits = self.gnn(img_embs, kwargs["adjacencies"])
        y_true = self.prepare_labels(kwargs[self.label_key], self.label_tokenizer)
        loss = self.get_loss_function()(logits, y_true)
        y_prob = self.prepare_y_prob(logits)
        return {
            "loss": loss,
            "y_prob": y_prob,
            "y_true": y_true,
        }


if __name__ == "__main__":
    
    from torchvision import transforms
    from pyhealth.datasets import COVID19CXRDataset
    from pyhealth.datasets import split_by_sample

    base_dataset = COVID19CXRDataset(
        root="/srv/local/data/COVID-19_Radiography_Dataset",
    )

    sample_dataset = base_dataset.set_task()

    transform = transforms.Compose([
        # transforms.Grayscale(),
        transforms.Lambda(lambda x: x if x.shape[0] == 3 else x.repeat(3, 1, 1)),
        transforms.Resize((224, 224)),
        transforms.Normalize(mean=[0.5862785803043838], std=[0.27950088968644304])
    ])


    def encode(sample):
        sample["path"] = transform(sample["path"])
        return sample

    sample_dataset.set_transform(encode)
    
    
    # Get Index of train, valid, test set
    train_index, val_index, test_index = split_by_sample(
        dataset=sample_dataset,
        ratios=[0.7, 0.1, 0.2],
        get_index = True
    )

    model = Graph_TorchvisionModel(
        dataset=sample_dataset,
        feature_keys=["path"],
        label_key="label",
        mode="multiclass",
        model_name="resnet18",
        # model_config={"weights": "DEFAULT"},
        model_config={},
        gnn_config={"input_dim": 256, "hidden_dim": 128},
    )

    graph = model.build_graph(sample_dataset, random = True)
    
    from pyhealth.datasets import split_by_sample

    # Get Index of train, valid, test set
    train_index, val_index, test_index = split_by_sample(
        dataset=sample_dataset,
        ratios=[0.7, 0.1, 0.2],
        get_index = True
    )
    
    # Define Sampler as Dataloader
    train_dataloader = NeighborSampler(sample_dataset, graph["edge_index"], node_idx=train_index, sizes=[15, 10], batch_size=64, shuffle=True, num_workers=12)
    # We sample all edges connected to target node for validation and test (Sizes = [-1, -1])
    valid_dataloader = NeighborSampler(sample_dataset, graph["edge_index"], node_idx=val_index, sizes=[-1, -1], batch_size=64, shuffle=False, num_workers=12)
    test_dataloader = NeighborSampler(sample_dataset, graph["edge_index"], node_idx=test_index, sizes=[-1, -1], batch_size=64, shuffle=False, num_workers=12)

    
    # train_dataloader = NeighborSampler(sample_dataset, graph["edge_index"], node_idx=train_index, sizes=[15, 10], batch_size=64, shuffle=True, num_workers=12)

    # data_graph_batch = next(iter(train_dataloader))

    # # try the model
    # ret = model(**data_graph_batch)
    # print(ret)

    # # try loss backward
    # ret["loss"].backward()
    
    from pyhealth.trainer import Trainer
    resnet_trainer = Trainer(model=model, device="cpu")
    resnet_trainer.train(
        train_dataloader=train_dataloader,
        val_dataloader=valid_dataloader,
        epochs=1,
        monitor="accuracy",
    )
    
    print(resnet_trainer.evaluate(test_dataloader))
    
    
    resnet_trainer.train(
        train_dataloader=train_dataloader,
        val_dataloader=valid_dataloader,
        epochs=1,
        monitor="accuracy",
    )



"""
Histogram Binning.

Implementation based on https://github.com/aigen/df-posthoc-calibration

"""
from typing import Dict

import numpy as np
import torch
from torch.utils.data import Subset

from pyhealth.calib.base_classes import PostHocCalibrator
from pyhealth.calib.utils import LogLoss, one_hot_np, prepare_numpy_dataset
from pyhealth.models import BaseModel

__all__ = ["HistogramBinning"]


def _nudge(matrix, delta):
    return (matrix + np.random.uniform(low=0, high=delta, size=matrix.shape)) / (
        1 + delta
    )


def _bin_points(scores, bin_edges):
    assert bin_edges is not None, "Bins have not been defined"
    scores = scores.squeeze()
    assert np.size(scores.shape) < 2, "scores should be a 1D vector or singleton"
    scores = np.reshape(scores, (scores.size, 1))
    bin_edges = np.reshape(bin_edges, (1, bin_edges.size))
    return np.sum(scores > bin_edges, axis=1)


def _get_uniform_mass_bins(probs, n_bins):
    assert probs.size >= n_bins, "Fewer points than bins"

    probs_sorted = np.sort(probs)

    # split probabilities into groups of approx equal size
    groups = np.array_split(probs_sorted, n_bins)
    bin_upper_edges = list()

    for cur_group in range(n_bins - 1):
        bin_upper_edges += [max(groups[cur_group])]
    bin_upper_edges += [np.inf]

    return np.array(bin_upper_edges)


class HB_binary(object):
    def __init__(self, n_bins=15):
        ### Hyperparameters
        self.delta = 1e-10
        self.n_bins = n_bins

        ### Parameters to be learnt
        self.bin_upper_edges = None
        self.mean_pred_values = None
        self.num_calibration_examples_in_bin = None

        ### Internal variables
        self.fitted = False

    def fit(self, y_score, y):
        assert self.n_bins is not None, "Number of bins has to be specified"
        y_score = y_score.squeeze()
        y = y.squeeze()
        assert y_score.size == y.size, "Check dimensions of input matrices"
        assert (
            y.size >= self.n_bins
        ), "Number of bins should be less than the number of calibration points"

        ### All required (hyper-)parameters have been passed correctly
        ### Uniform-mass binning/histogram binning code starts below

        # delta-randomization
        y_score = _nudge(y_score, self.delta)

        # compute uniform-mass-bins using calibration data
        self.bin_upper_edges = _get_uniform_mass_bins(y_score, self.n_bins)

        # assign calibration data to bins
        bin_assignment = _bin_points(y_score, self.bin_upper_edges)

        # compute bias of each bin
        self.num_calibration_examples_in_bin = np.zeros([self.n_bins, 1])
        self.mean_pred_values = np.empty(self.n_bins)
        for i in range(self.n_bins):
            bin_idx = bin_assignment == i
            self.num_calibration_examples_in_bin[i] = sum(bin_idx)

            # nudge performs delta-randomization
            if sum(bin_idx) > 0:
                self.mean_pred_values[i] = _nudge(y[bin_idx].mean(), self.delta)
            else:
                self.mean_pred_values[i] = _nudge(0.5, self.delta)

        # check that my code is correct
        assert np.sum(self.num_calibration_examples_in_bin) == y.size

        # histogram binning done
        self.fitted = True
        return self

    def predict_proba(self, y_score):
        assert self.fitted is True, "Call HB_binary.fit() first"
        y_score = y_score.squeeze()

        # delta-randomization
        y_score = _nudge(y_score, self.delta)

        # assign test data to bins
        y_bins = _bin_points(y_score, self.bin_upper_edges)

        # get calibrated predicted probabilities
        y_pred_prob = self.mean_pred_values[y_bins]
        return y_pred_prob


# HB_toplabel is removed as it has no use in our tasks


class HistogramBinning(PostHocCalibrator):
    """Histogram Binning

    Histogram binning amounts to creating bins and computing the accuracy
    for each bin using the calibration dataset, and then predicting such
    at test time. For multilabel/binary/multiclass classification tasks,
    we calibrate each class independently following [1]. Users could choose
    to renormalize the probability scores for multiclass tasks so they sum
    to 1.


    Paper:

        [1] Gupta, Chirag, and Aaditya Ramdas.
        "Top-label calibration and multiclass-to-binary reductions."
        ICLR 2022.

        [2] Zadrozny, Bianca, and Charles Elkan.
        "Learning and making decisions when costs and probabilities are both unknown."
        In Proceedings of the seventh ACM SIGKDD international conference on Knowledge
        discovery and data mining, pp. 204-213. 2001.

    :param model: A trained base model.
    :type model: BaseModel

    Examples:
        >>> from pyhealth.datasets import ISRUCDataset, get_dataloader, split_by_patient
        >>> from pyhealth.models import SparcNet
        >>> from pyhealth.tasks import sleep_staging_isruc_fn
        >>> from pyhealth.calib.calibration import HistogramBinning
        >>> sleep_ds = ISRUCDataset("/srv/scratch1/data/ISRUC-I").set_task(sleep_staging_isruc_fn)
        >>> train_data, val_data, test_data = split_by_patient(sleep_ds, [0.6, 0.2, 0.2])
        >>> model = SparcNet(dataset=sleep_ds, feature_keys=["signal"],
        ...     label_key="label", mode="multiclass")
        >>> # ... Train the model here ...
        >>> # Calibrate
        >>> cal_model = HistogramBinning(model)
        >>> cal_model.calibrate(cal_dataset=val_data)
        >>> # Evaluate
        >>> from pyhealth.trainer import Trainer
        >>> test_dl = get_dataloader(test_data, batch_size=32, shuffle=False)
        >>> print(Trainer(model=cal_model, metrics=['cwECEt_adapt', 'accuracy']).evaluate(test_dl))
        {'accuracy': 0.7189072348464207, 'cwECEt_adapt': 0.04455814993598299}
    """

    def __init__(self, model: BaseModel, debug=False, **kwargs) -> None:
        super().__init__(model, **kwargs)
        self.mode = self.model.mode
        for param in model.parameters():
            param.requires_grad = False

        self.model.eval()
        self.device = model.device
        self.debug = debug

        self.num_classes = None
        self.calib = None

    def calibrate(self, cal_dataset: Subset, nbins: int = 15):
        """Calibrate the base model using a calibration dataset.

        :param cal_dataset: Calibration set.
        :type cal_dataset: Subset
        :param nbins: number of bins to use, defaults to 15
        :type nbins: int, optional
        """
        _cal_data = prepare_numpy_dataset(
            self.model, cal_dataset, ["y_true", "y_prob"], debug=self.debug
        )
        if self.num_classes is None:
            self.num_classes = _cal_data["y_prob"].shape[1]
        if self.mode == "binary":
            self.calib = [
                HB_binary(nbins).fit(
                    _cal_data["y_prob"][:, 0], _cal_data["y_true"][:, 0]
                )
            ]
        else:
            self.calib = []
            y_true = _cal_data["y_true"]
            if len(y_true.shape) == 1:
                y_true = one_hot_np(y_true, self.num_classes)
            for k in range(self.num_classes):
                self.calib.append(
                    HB_binary().fit(_cal_data["y_prob"][:, k], y_true[:, k])
                )

    def forward(self, normalization="sum", **kwargs) -> Dict[str, torch.Tensor]:
        """Forward propagation (just like the original model).

        :param normalization: how to normalize the calibrated probability.
            Defaults to 'sum' (and only 'sum' is supported for now).
        :type normalization: str, optional
        :param **kwargs: Additional arguments to the base model.

        :return:  A dictionary with all results from the base model, with the following modified:

            ``y_prob``: calibrated predicted probabilities.
            ``loss``: Cross entropy loss  with the new y_prob.
        :rtype: Dict[str, torch.Tensor]
        """
        assert normalization is None or normalization == "sum"
        ret = self.model(**kwargs)
        y_prob = ret["y_prob"].cpu().numpy()
        for k in range(self.num_classes):
            y_prob[:, k] = self.calib[k].predict_proba(y_prob[:, k])
        if normalization == "sum" and self.mode == "multiclass":
            y_prob = y_prob / y_prob.sum(1)[:, np.newaxis]
        ret["y_prob"] = torch.tensor(y_prob, dtype=torch.float, device=self.device)
        if self.mode == "multiclass":
            criterion = LogLoss()
        else:
            criterion = torch.nn.functional.binary_cross_entropy
        ret["loss"] = criterion(ret["y_prob"], ret["y_true"])
        ret.pop("logit")
        return ret


if __name__ == "__main__":
    from pyhealth.datasets import ISRUCDataset, get_dataloader, split_by_patient
    from pyhealth.models import SparcNet
    from pyhealth.tasks import sleep_staging_isruc_fn
    from pyhealth.calib.calibration import HistogramBinning

    sleep_ds = ISRUCDataset(
        root="/srv/local/data/trash/",
        dev=True,
    ).set_task(sleep_staging_isruc_fn)
    train_data, val_data, test_data = split_by_patient(sleep_ds, [0.6, 0.2, 0.2])
    model = SparcNet(
        dataset=sleep_ds,
        feature_keys=["signal"],
        label_key="label",
        mode="multiclass",
    )
    # ... Train the model here ...
    # Calibrate
    cal_model = HistogramBinning(model)
    cal_model.calibrate(cal_dataset=val_data)
    # Evaluate
    from pyhealth.trainer import Trainer

    test_dl = get_dataloader(test_data, batch_size=32, shuffle=False)
    print(
        Trainer(model=cal_model, metrics=["cwECEt_adapt", "accuracy"]).evaluate(test_dl)
    )
